#include <msp/core/algorithm.h>
#include <msp/core/maputils.h>
#include "error.h"
#include "renderbuffer.h"
#include "sequence.h"
#include "sequencebuilder.h"
#include "sequencetemplate.h"

using namespace std;

namespace Msp {
namespace GL {

SequenceBuilder::SequenceBuilder(const SequenceTemplate &t):
	tmpl(t)
{
	const vector<SequenceTemplate::Step> &steps = tmpl.get_steps();
	for(vector<SequenceTemplate::Step>::const_iterator i=steps.begin(); i!=steps.end(); ++i)
		renderables[i->slot_name] = 0;
	const vector<SequenceTemplate::PostProcessor> &postprocs = tmpl.get_postprocessors();
	for(SequenceTemplate::PostProcessorArray::const_iterator i=postprocs.begin(); i!=postprocs.end(); ++i)
		if(!i->slot_name.empty())
			postprocessors[i->slot_name] = 0;
}

void SequenceBuilder::set_renderable(const string &name, Renderable &rend)
{
	get_item(renderables, name) = &rend;
}

void SequenceBuilder::set_postprocessor(const string &name, PostProcessor &pproc)
{
	get_item(postprocessors, name) = &pproc;
}

void SequenceBuilder::build(Sequence &sequence) const
{
	sequence.set_hdr(tmpl.get_hdr());
	sequence.set_alpha(tmpl.get_alpha());
	unsigned samples = min(tmpl.get_maximum_multisample(), Renderbuffer::get_max_samples());
	if(samples<tmpl.get_required_multisample())
		throw invalid_operation("SequenceBuilder::build");

	sequence.set_multisample(samples);

	const vector<SequenceTemplate::Step> &steps = tmpl.get_steps();
	for(vector<SequenceTemplate::Step>::const_iterator i=steps.begin(); i!=steps.end(); ++i)
	{
		Renderable *renderable = get_item(renderables, i->slot_name);
		if(!renderable)
			continue;

		Sequence::Step &step = sequence.add_step(i->tag, *renderable);
		step.set_blend(i->blend.get());
		step.set_depth_test(i->depth_test.get());
		step.set_lighting(i->lighting.get());
	}

	const SequenceTemplate::PostProcessorArray &postprocs = tmpl.get_postprocessors();
	for(SequenceTemplate::PostProcessorArray::const_iterator i=postprocs.begin(); i!=postprocs.end(); ++i)
	{
		PostProcessor *proc = 0;
		if(!i->slot_name.empty())
			proc = get_item(postprocessors, i->slot_name);
		if(proc)
			sequence.add_postprocessor(*proc);
		else if(i->postprocessor_template)
		{
			proc = i->postprocessor_template->create(tmpl.get_resources(), sequence.get_width(), sequence.get_height());
			if(proc)
				sequence.add_postprocessor_owned(proc);
		}
	}
}

Sequence *SequenceBuilder::build(unsigned w, unsigned h) const
{
	RefPtr<Sequence> sequence = new Sequence(w, h);
	build(*sequence);
	return sequence.release();
}

Sequence *SequenceBuilder::build(const View &view) const
{
	RefPtr<Sequence> sequence = new Sequence(view);
	build(*sequence);
	return sequence.release();
}

Sequence *SequenceBuilder::build(const Framebuffer &fbo) const
{
	RefPtr<Sequence> sequence = new Sequence(fbo);
	build(*sequence);
	return sequence.release();
}

} // namespace GL
} // namespace Msp
