#include <msp/gl/extensions/arb_direct_state_access.h>
#include <msp/gl/extensions/arb_texture_swizzle.h>
#include <msp/gl/extensions/ext_framebuffer_object.h>
#include <msp/io/memory.h>
#include "bindable.h"
#include "error.h"
#include "resourcemanager.h"
#include "resources.h"
#include "texture.h"
#include "texunit.h"

using namespace std;

namespace Msp {
namespace GL {

int Texture::swizzle_orders[] =
{
	GL_RED, GL_GREEN, GL_BLUE, GL_ALPHA,
	GL_RED, GL_RED, GL_RED, GL_ONE,
	GL_RED, GL_RED, GL_RED, GL_GREEN,
	GL_BLUE, GL_GREEN, GL_RED, GL_ALPHA
};

Texture::Texture(GLenum t, ResourceManager *m):
	id(0),
	target(t),
	format(RGB8),
	storage_fmt(RGB8),
	swizzle(NO_SWIZZLE),
	use_srgb_format(false),
	auto_gen_mipmap(false),
	default_sampler(*this)
{
	if(m)
		set_manager(m);
	else if(ARB_direct_state_access)
		glCreateTextures(target, 1, &id);
	else
		glGenTextures(1, &id);
}

Texture::~Texture()
{
	while(TexUnit *unit = TexUnit::find_unit(this))
		unbind_from(unit->get_index());

	if(id)
		glDeleteTextures(1, &id);
}

void Texture::set_format(PixelFormat fmt)
{
	PixelComponents comp = get_components(fmt);
	PixelComponents st_comp = comp;
	FormatSwizzle swiz = NO_SWIZZLE;
	switch(comp)
	{
	case LUMINANCE:
		st_comp = RED;
		swiz = R_TO_LUMINANCE;
		break;
	case LUMINANCE_ALPHA:
		st_comp = RG;
		swiz = RG_TO_LUMINANCE_ALPHA;
		break;
	case BGR:
		st_comp = RGB;
		swiz = RGB_TO_BGR;
		break;
	case BGRA:
		st_comp = RGBA;
		swiz = RGB_TO_BGR;
		break;
	default:;
	}

	PixelFormat st_fmt = make_pixelformat(st_comp, get_component_type(fmt), is_srgb(fmt));
	require_pixelformat(st_fmt);
	if(swiz!=NO_SWIZZLE)
		static Require _req(ARB_texture_swizzle);

	format = fmt;
	storage_fmt = st_fmt;
	swizzle = swiz;
}

void Texture::apply_swizzle()
{
	if(swizzle==NO_SWIZZLE)
		return;

	if(get_gl_api()==OPENGL_ES2)
	{
		set_parameter_i(GL_TEXTURE_SWIZZLE_R, swizzle_orders[swizzle*4]);
		set_parameter_i(GL_TEXTURE_SWIZZLE_G, swizzle_orders[swizzle*4+1]);
		set_parameter_i(GL_TEXTURE_SWIZZLE_B, swizzle_orders[swizzle*4+2]);
		set_parameter_i(GL_TEXTURE_SWIZZLE_A, swizzle_orders[swizzle*4+3]);
	}
	else
	{
		if(ARB_direct_state_access)
			glTextureParameteriv(id, GL_TEXTURE_SWIZZLE_RGBA, swizzle_orders+swizzle*4);
		else
			glTexParameteriv(target, GL_TEXTURE_SWIZZLE_RGBA, swizzle_orders+swizzle*4);
	}
}

void Texture::set_parameter_i(GLenum param, int value) const
{
	if(ARB_direct_state_access)
		glTextureParameteri(id, param, value);
	else
		glTexParameteri(target, param, value);
}

void Texture::set_min_filter(TextureFilter f)
{
	default_sampler.set_min_filter(f);
}

void Texture::set_mag_filter(TextureFilter f)
{
	default_sampler.set_mag_filter(f);
}

void Texture::set_filter(TextureFilter f)
{
	default_sampler.set_filter(f);
}

void Texture::set_max_anisotropy(float a)
{
	default_sampler.set_max_anisotropy(a);
}

void Texture::set_wrap(TextureWrap w)
{
	default_sampler.set_wrap(w);
}

void Texture::set_wrap_s(TextureWrap w)
{
	default_sampler.set_wrap_s(w);
}

void Texture::set_wrap_t(TextureWrap w)
{
	default_sampler.set_wrap_t(w);
}

void Texture::set_wrap_r(TextureWrap w)
{
	default_sampler.set_wrap_r(w);
}

bool Texture::can_generate_mipmap()
{
	return EXT_framebuffer_object;
}

void Texture::generate_mipmap()
{
	// glGenerateMipmap is defined here
	static Require _req(EXT_framebuffer_object);

	if(ARB_direct_state_access)
		glGenerateTextureMipmap(id);
	else
	{
		BindRestore _bind(this);
		glGenerateMipmap(target);
	}
}

void Texture::set_auto_generate_mipmap(bool gm)
{
	if(gm)
		static Require _req(EXT_framebuffer_object);

	auto_gen_mipmap = gm;
}

void Texture::set_compare_enabled(bool c)
{
	if(c)
		default_sampler.set_compare(default_sampler.get_compare_function());
	else
		default_sampler.disable_compare();
}

void Texture::set_compare_func(Predicate f)
{
	default_sampler.set_compare(f);
}

void Texture::load_image(const string &fn, bool)
{
	load_image(fn, 0U);
}

void Texture::load_image(const string &fn, unsigned lv)
{
	Graphics::Image img;
	img.load_file(fn);

	image(img, lv);
}

void Texture::image(const Graphics::Image &img, bool)
{
	image(img, 0U);
}

void Texture::bind_to(unsigned i) const
{
	if(!id)
	{
		if(manager)
			manager->resource_used(*this);
		if(!id)
		{
			unbind_from(i);
			return;
		}
	}

	TexUnit &unit = TexUnit::get_unit(i);
	const Texture *cur = unit.get_texture();
	if(unit.set_texture(this))
	{
		if(manager)
			manager->resource_used(*this);

		if(ARB_direct_state_access)
			glBindTextureUnit(i, id);
		else
		{
			unit.bind();
			glBindTexture(target, id);
		}

		if(!unit.get_sampler() || unit.get_sampler()==&cur->default_sampler)
			default_sampler.bind_to(i);
	}
}

const Texture *Texture::current(unsigned i)
{
	return TexUnit::get_unit(i).get_texture();
}

void Texture::unbind_from(unsigned i)
{
	TexUnit &unit = TexUnit::get_unit(i);
	const Texture *cur = unit.get_texture();
	if(unit.set_texture(0))
	{
		if(ARB_direct_state_access)
			glBindTextureUnit(i, 0);
		else
		{
			unit.bind();
			glBindTexture(cur->target, 0);
		}

		if(unit.get_sampler()==&cur->default_sampler)
			Sampler::unbind_from(i);
	}
}


Texture::Loader::Loader(Texture &t):
	DataFile::CollectionObjectLoader<Texture>(t, 0)
{
	init();
}

Texture::Loader::Loader(Texture &t, Collection &c):
	DataFile::CollectionObjectLoader<Texture>(t, &c)
{
	init();
}

void Texture::Loader::init()
{
	levels = 0;

	add("external_image", &Loader::external_image);
	add("external_image_srgb", &Loader::external_image_srgb);
	add("filter", &Loader::filter);
	add("generate_mipmap", &Loader::generate_mipmap);
	add("image_data", &Loader::image_data);
	add("mag_filter", &Loader::mag_filter);
	add("max_anisotropy", &Loader::max_anisotropy);
	add("min_filter", &Loader::min_filter);
	add("mipmap_levels", &Loader::mipmap_levels);
	add("sampler",    &Loader::sampler);
	add("wrap",       &Loader::wrap);
	add("wrap_r",     &Loader::wrap_r);
	add("wrap_s",     &Loader::wrap_s);
	add("wrap_t",     &Loader::wrap_t);
}

unsigned Texture::Loader::get_levels() const
{
	return (is_mipmapped(obj.default_sampler.get_min_filter()) ? levels : 1);
}

#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wdeprecated-declarations"
void Texture::Loader::load_external_image(Graphics::Image &img, const string &fn)
{
	RefPtr<IO::Seekable> io = get_collection().open_raw(fn);
	if(!io)
		throw IO::file_not_found(fn);
	img.load_io(*io);
}

void Texture::Loader::external_image(const string &fn)
{
	obj.use_srgb_format = false;
	external_image_common(fn);
}

void Texture::Loader::external_image_srgb(const string &fn)
{
	obj.use_srgb_format = true;
	external_image_common(fn);
}

void Texture::Loader::external_image_common(const string &fn)
{
	if(obj.manager)
		obj.manager->set_resource_location(obj, get_collection(), fn);
	else
	{
		Graphics::Image img;
		load_external_image(img, fn);
		obj.image(img, get_levels());
	}
}

void Texture::Loader::filter(TextureFilter f)
{
	obj.set_filter(f);
}

void Texture::Loader::generate_mipmap(bool gm)
{
	obj.set_auto_generate_mipmap(gm);
}

void Texture::Loader::image_data(const string &data)
{
	Graphics::Image img;
	IO::Memory mem(data.data(), data.size());
	img.load_io(mem);

	obj.image(img, get_levels());
}

void Texture::Loader::mag_filter(TextureFilter f)
{
	obj.set_mag_filter(f);
}

void Texture::Loader::max_anisotropy(float a)
{
	obj.set_max_anisotropy(a);
}

void Texture::Loader::min_filter(TextureFilter f)
{
	obj.set_min_filter(f);
}

void Texture::Loader::mipmap_levels(unsigned l)
{
	levels = l;
}

void Texture::Loader::sampler()
{
	load_sub(obj.default_sampler);
}

void Texture::Loader::wrap(TextureWrap w)
{
	obj.set_wrap(w);
}

void Texture::Loader::wrap_r(TextureWrap w)
{
	obj.set_wrap_r(w);
}

void Texture::Loader::wrap_s(TextureWrap w)
{
	obj.set_wrap_s(w);
}

void Texture::Loader::wrap_t(TextureWrap w)
{
	obj.set_wrap_t(w);
}
#pragma GCC diagnostic pop

} // namespace GL
} // namespace Msp
