#include <msp/core/raii.h>
#include <msp/gl/extensions/arb_direct_state_access.h>
#include <msp/gl/extensions/arb_texture_storage.h>
#include <msp/graphics/imageloader.h>
#include "bindable.h"
#include "buffer.h"
#include "error.h"
#include "pixelstore.h"
#include "resources.h"
#include "texture2d.h"

using namespace std;

namespace Msp {
namespace GL {

class Texture2D::AsyncLoader: public Resource::AsyncLoader
{
private:
	Texture2D &texture;
	IO::Seekable &io;
	Buffer pixel_buffer;
	char *mapped_address;
	Graphics::Image image;
	Graphics::ImageLoader *img_loader;
	unsigned n_bytes;
	int phase;

public:
	AsyncLoader(Texture2D &, IO::Seekable &);
	~AsyncLoader();

	virtual bool needs_sync() const;
	virtual bool process();
};


Texture2D::Texture2D(ResourceManager *m):
	Texture(GL_TEXTURE_2D, m),
	width(0),
	height(0),
	allocated(0)
{ }

Texture2D::~Texture2D()
{
	set_manager(0);
}

void Texture2D::storage(PixelFormat fmt, unsigned wd, unsigned ht, unsigned lv)
{
	if(width>0)
		throw invalid_operation("Texture2D::storage");
	if(wd==0 || ht==0)
		throw invalid_argument("Texture2D::storage");

	set_format(fmt);
	width = wd;
	height = ht;
	levels = get_n_levels();
	if(lv>0)
		levels = min(levels, lv);
}

void Texture2D::allocate(unsigned level)
{
	if(width==0 || height==0)
		throw invalid_operation("Texture2D::allocate");
	if(level>=levels)
		throw invalid_argument("Texture2D::allocate");
	if(allocated&(1<<level))
		return;

	if(ARB_texture_storage)
	{
		Conditional<BindRestore> _bind(!ARB_direct_state_access, this);
		if(ARB_direct_state_access)
			glTextureStorage2D(id, levels, storage_fmt, width, height);
		else
			glTexStorage2D(target, levels, storage_fmt, width, height);
		apply_swizzle();
		allocated |= (1<<levels)-1;
	}
	else
		image(level, 0);
}

void Texture2D::image(unsigned level, const void *data)
{
	if(width==0 || height==0)
		throw invalid_operation("Texture2D::image");

	unsigned w = width;
	unsigned h = height;
	get_level_size(level, w, h);

	if(ARB_texture_storage)
		return sub_image(level, 0, 0, w, h, data);

	BindRestore _bind(this);

	if(!allocated)
	{
		glTexParameteri(target, GL_TEXTURE_MAX_LEVEL, levels-1);
		apply_swizzle();
	}

	PixelComponents comp = get_components(storage_fmt);
	GLenum type = get_gl_type(get_component_type(storage_fmt));
	glTexImage2D(target, level, storage_fmt, w, h, 0, comp, type, data);

	allocated |= 1<<level;
	if(auto_gen_mipmap && level==0)
	{
		generate_mipmap();
		allocated |= (1<<levels)-1;
	}
}

void Texture2D::image(unsigned level, PixelComponents comp, DataType type, const void *data)
{
	if(comp!=get_components(format) || type!=get_component_type(format))
		throw incompatible_data("Texture2D::image");
	image(level, data);
}

void Texture2D::sub_image(unsigned level, int x, int y, unsigned wd, unsigned ht, const void *data)
{
	if(width==0 || height==0)
		throw invalid_operation("Texture2D::sub_image");

	Conditional<BindRestore> _bind(!ARB_direct_state_access, this);
	allocate(level);

	PixelComponents comp = get_components(storage_fmt);
	GLenum type = get_gl_type(get_component_type(storage_fmt));
	if(ARB_direct_state_access)
		glTextureSubImage2D(id, level, x, y, wd, ht, comp, type, data);
	else
		glTexSubImage2D(target, level, x, y, wd, ht, comp, type, data);

	if(auto_gen_mipmap && level==0)
		generate_mipmap();
}

void Texture2D::sub_image(unsigned level, int x, int y, unsigned wd, unsigned ht, PixelComponents comp, DataType type, const void *data)
{
	if(comp!=get_components(format) || type!=get_component_type(format))
		throw incompatible_data("Texture2D::sub_image");
	sub_image(level, x, y, wd, ht, data);
}

void Texture2D::image(const Graphics::Image &img, unsigned lv)
{
	image(img, lv, false);
}

void Texture2D::image(const Graphics::Image &img, unsigned lv, bool from_buffer)
{
	unsigned w = img.get_width();
	unsigned h = img.get_height();
	PixelFormat fmt = pixelformat_from_image(img);
	if(width==0)
		storage(make_pixelformat(get_components(fmt), get_component_type(fmt), use_srgb_format), w, h, lv);
	else if(w!=width || h!=height || (lv && lv!=levels))
		throw incompatible_data("Texture2D::image");

	PixelStore pstore = PixelStore::from_image(img);
	BindRestore _bind_ps(pstore);

	image(0, from_buffer ? 0 : img.get_pixels());
}

unsigned Texture2D::get_n_levels() const
{
	unsigned n = 0;
	for(unsigned s=max(width, height); s; s>>=1, ++n) ;
	return n;
}

void Texture2D::get_level_size(unsigned level, unsigned &w, unsigned &h) const
{
	w >>= level;
	h >>= level;

	if(!w && h)
		w = 1;
	else if(!h && w)
		h = 1;
}

Resource::AsyncLoader *Texture2D::load(IO::Seekable &io, const Resources *)
{
	AsyncLoader *ldr = new AsyncLoader(*this, io);
	return ldr;
}

UInt64 Texture2D::get_data_size() const
{
	return id ? width*height*get_pixel_size(format) : 0;
}

void Texture2D::unload()
{
	glDeleteTextures(1, &id);
	id = 0;
	allocated = 0;
	default_sampler.unload();
}


Texture2D::Loader::Loader(Texture2D &t):
	DataFile::DerivedObjectLoader<Texture2D, Texture::Loader>(t)
{
	init();
}

Texture2D::Loader::Loader(Texture2D &t, Collection &c):
	DataFile::DerivedObjectLoader<Texture2D, Texture::Loader>(t, c)
{
	init();
}

void Texture2D::Loader::init()
{
	add("raw_data", &Loader::raw_data);
	add("storage", &Loader::storage);
	add("storage", &Loader::storage_levels);
}

void Texture2D::Loader::raw_data(const string &data)
{
	obj.image(0, data.data());
}

void Texture2D::Loader::storage(PixelFormat fmt, unsigned w, unsigned h)
{
	obj.storage(fmt, w, h);
}

void Texture2D::Loader::storage_levels(PixelFormat fmt, unsigned w, unsigned h, unsigned l)
{
	obj.storage(fmt, w, h, l);
}


Texture2D::AsyncLoader::AsyncLoader(Texture2D &t, IO::Seekable &i):
	texture(t),
	io(i),
	pixel_buffer(PIXEL_UNPACK_BUFFER),
	mapped_address(0),
	img_loader(Graphics::ImageLoader::open_io(io)),
	phase(0)
{ }

Texture2D::AsyncLoader::~AsyncLoader()
{
	if(mapped_address)
		pixel_buffer.unmap();
	delete img_loader;
}

bool Texture2D::AsyncLoader::needs_sync() const
{
	return phase%2;
}

bool Texture2D::AsyncLoader::process()
{
	if(phase==0)
	{
		image.load_headers(*img_loader);
		n_bytes = image.get_stride()*image.get_height();
	}
	else if(phase==1)
	{
		pixel_buffer.storage(n_bytes);
		mapped_address = reinterpret_cast<char *>(pixel_buffer.map());
	}
	else if(phase==2)
		image.load_into(*img_loader, mapped_address);
	else if(phase==3)
	{
		Bind _bind_buf(pixel_buffer, PIXEL_UNPACK_BUFFER);
		mapped_address = 0;
		if(!pixel_buffer.unmap())
		{
			phase = 1;
			return false;
		}

		if(!texture.id)
		{
			if(ARB_direct_state_access)
				glCreateTextures(texture.target, 1, &texture.id);
			else
				glGenTextures(1, &texture.id);
		}
		texture.image(image, 0, true);
	}

	++phase;
	return phase>3;
}

} // namespace GL
} // namespace Msp
