#ifndef MSP_GL_MATERIAL_H_
#define MSP_GL_MATERIAL_H_

#include <msp/core/typeregistry.h>
#include <msp/datafile/collection.h>
#include <msp/datafile/objectloader.h>
#include "color.h"
#include "programdata.h"
#include "texture.h"

namespace Msp {
namespace GL {

class Texturing;

class Material
{
private:
	class Loader: public DataFile::CollectionObjectLoader<Material>
	{
	protected:
		Loader(Material &);
		Loader(Material &, Collection &);

		virtual void init_actions();

	private:
		void sampler(const std::string &);
	};

protected:
	template<typename T>
	struct Property
	{
		T value;
		const Texture *texture;

		Property(): value(T()), texture(0) { }
	};

	template<typename C>
	class PropertyLoader: public DataFile::DerivedObjectLoader<Material, Loader>
	{
	protected:
		PropertyLoader(C &m): DerivedObjectLoader<Material, Loader>(m) { }
		PropertyLoader(C &m, Collection &c): DerivedObjectLoader<Material, Loader>(m, c) { }

		void add_property(const std::string &, void (C::*)(float), void (C::*)(const Texture *));
		void add_property(const std::string &, void (C::*)(const Color &), void (C::*)(const Texture *), bool);
		void add_property(const std::string &, void (C::*)(const Texture *));

		void property_value_scalar(void (C::*)(float), float);
		void property_value_rgb(void (C::*)(const Color &), float, float, float);
		void property_value_rgba(void (C::*)(const Color &), float, float, float, float);
		void property_value_srgb(void (C::*)(const Color &), float, float, float);
		void property_value_srgb_alpha(void (C::*)(const Color &), float, float, float, float);
		void property_texture(void (C::*)(const Texture *), const std::string &);
	};

public:
	class GenericLoader: public DataFile::Loader
	{
	private:
		template<typename T>
		struct CreateMaterial
		{
			void operator()(const std::string &, GenericLoader &) const;
		};

		DataFile::Collection *coll;
		Material *material;
		Loader *mat_loader;

		static ActionMap shared_actions;

	public:
		GenericLoader(DataFile::Collection * = 0);
		~GenericLoader();

		Material *get_material() { Material *m = material; material = 0; return m; }
	private:
		virtual void init_actions();

		void type(const DataFile::Symbol &);

		friend class Material;
	};

private:
	typedef TypeRegistry<GenericLoader::CreateMaterial, GenericLoader &> MaterialRegistry;

protected:
	const Sampler *sampler;
	ProgramData shdata;

	Material(): sampler(0) { }
public:
	virtual ~Material() { }

	virtual const Program *create_compatible_shader(DataFile::Collection &) const;
protected:
	virtual void fill_program_info(std::string &, std::map<std::string, int> &) const = 0;

public:
	/** Returns the uniforms for the material. */
	const ProgramData &get_shader_data() const { return shdata; }

protected:
	DEPRECATED void attach_texture_to(const Texture *, Texturing &, ProgramData &, const std::string &) const;
public:
	DEPRECATED virtual void attach_textures_to(Texturing &, ProgramData &) const = 0;

	virtual const Tag *get_texture_tags() const = 0;
	virtual const Texture *get_texture(Tag) const = 0;
	const Sampler *get_sampler() const { return sampler; }

	template<typename T>
	static void register_type(const std::string &);
private:
	static MaterialRegistry &get_material_registry();
};

template<typename T>
void Material::register_type(const std::string &kw)
{
	get_material_registry().register_type<T>(kw);
}


template<typename C>
void Material::PropertyLoader<C>::add_property(const std::string &kw, void (C::*set_value)(float), void (C::*set_texture)(const Texture *))
{
	add(kw, &PropertyLoader<C>::property_value_scalar, set_value);
	if(set_texture)
		add(kw+"_map", &PropertyLoader<C>::property_texture, set_texture);
}

template<typename C>
void Material::PropertyLoader<C>::add_property(const std::string &kw, void (C::*set_value)(const Color &), void (C::*set_texture)(const Texture *), bool allow_alpha)
{
	add(kw, &PropertyLoader<C>::property_value_rgb, set_value);
	add(kw+"_srgb", &PropertyLoader<C>::property_value_srgb, set_value);
	if(allow_alpha)
	{
		add(kw, &PropertyLoader<C>::property_value_rgba, set_value);
		add(kw+"_srgb", &PropertyLoader<C>::property_value_srgb_alpha, set_value);
	}
	if(set_texture)
		add(kw+"_map", &PropertyLoader<C>::property_texture, set_texture);
}

template<typename C>
void Material::PropertyLoader<C>::add_property(const std::string &kw, void (C::*set_texture)(const Texture *))
{
	add(kw+"_map", &PropertyLoader<C>::property_texture, set_texture);
}

template<typename C>
void Material::PropertyLoader<C>::property_value_scalar(void (C::*set_value)(float), float value)
{
	(static_cast<C &>(obj).*set_value)(value);
}

template<typename C>
void Material::PropertyLoader<C>::property_value_rgb(void (C::*set_value)(const Color &), float r, float g, float b)
{
	(static_cast<C &>(obj).*set_value)(Color(r, g, b));
}

template<typename C>
void Material::PropertyLoader<C>::property_value_rgba(void (C::*set_value)(const Color &), float r, float g, float b, float a)
{
	(static_cast<C &>(obj).*set_value)(Color(r, g, b, a));
}

template<typename C>
void Material::PropertyLoader<C>::property_value_srgb(void (C::*set_value)(const Color &), float r, float g, float b)
{
	(static_cast<C &>(obj).*set_value)(Color(r, g, b).to_linear());
}

template<typename C>
void Material::PropertyLoader<C>::property_value_srgb_alpha(void (C::*set_value)(const Color &), float r, float g, float b, float a)
{
	(static_cast<C &>(obj).*set_value)(Color(r, g, b, a).to_linear());
}

template<typename C>
void Material::PropertyLoader<C>::property_texture(void (C::*set_texture)(const Texture *), const std::string &name)
{
	/* The static_cast around get_collection is needed because otherwise Android
	SDK's g++ 4.9 fails to parse get<Texture> as a template function call */
	(static_cast<C &>(obj).*set_texture)(&static_cast<Collection &>(get_collection()).get<Texture>(name));
}


template<typename T>
void Material::GenericLoader::CreateMaterial<T>::operator()(const std::string &, GenericLoader &ldr) const
{
	if(ldr.material)
		throw std::logic_error("Material type was already specified");

	T *mat = new T;
	ldr.material = mat;
	if(ldr.coll)
		ldr.mat_loader = new typename T::Loader(*mat, *ldr.coll);
	else
		ldr.mat_loader = new typename T::Loader(*mat);
	ldr.add_auxiliary_loader(*ldr.mat_loader);
}

} // namespace GL
} // namespace Msp

#endif
